@@ -490,19 +490,6 @@ class LightningRpc(UnixDomainSocketRpc):
         }
         return self.call("addgossip", payload)
 
-    def autocleaninvoice(self, cycle_seconds=None, expired_by=None):
-        """
-        Sets up automatic cleaning of expired invoices. {cycle_seconds} sets
-        the cleaning frequency in seconds (defaults to 3600) and {expired_by}
-        sets the minimum time an invoice should have been expired for to be
-        cleaned in seconds (defaults to 86400).
-        """
-        payload = {
-            "cycle_seconds": cycle_seconds,
-            "expired_by": expired_by
-        }
-        return self.call("autocleaninvoice", payload)
-
     def autoclean_status(self, subsystem=None):
         """
         Print status of autocleaning (optionally, just for {subsystem}).
@@ -607,15 +594,6 @@ class LightningRpc(UnixDomainSocketRpc):
         }
         return self.call("deldatastore", payload)
 
-    def delexpiredinvoice(self, maxexpirytime=None):
-        """
-        Delete all invoices that have expired on or before the given {maxexpirytime}.
-        """
-        payload = {
-            "maxexpirytime": maxexpirytime
-        }
-        return self.call("delexpiredinvoice", payload)
-
     def delinvoice(self, label, status, desconly=None):
         """
         Delete unpaid invoice {label} with {status} (or, with {desconly} true, remove its description).
@@ -1085,7 +1063,7 @@ class LightningRpc(UnixDomainSocketRpc):
     def pay(self, bolt11, amount_msat=None, label=None, riskfactor=None,
             maxfeepercent=None, retry_for=None,
             maxdelay=None, exemptfee=None, localinvreqid=None, exclude=None,
-            maxfee=None, description=None):
+            maxfee=None, description=None, partial_msat=None):
         """
         Send payment specified by {bolt11} with {amount_msat}
         (ignored if {bolt11} has an amount), optional {label}
@@ -1104,6 +1082,7 @@ class LightningRpc(UnixDomainSocketRpc):
             "exclude": exclude,
             "maxfee": maxfee,
             "description": description,
+            "partial_msat": partial_msat,
         }
         return self.call("pay", payload)
 
